/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PHIDAL_SymmetrizeMatrix.c                                  */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "phidal_ordering.h"


/*#define SPKIT_F*/
/* utilisation du code fortran en complexe ... c propre ;) */
#undef SPKIT_F /** On utilise faplb et csrcsc en C (int64) **/


#ifdef  SPKIT_F
#ifdef TYPE_COMPLEX
#define FFaplb Faplbc
#define FFcsrcsc Fcsrcscc
#else 
#define FFaplb Faplb
#define FFcsrcsc Fcsrcsc
#endif
#endif


void PHIDAL_SymmetrizeMatrix(flag_t job, flag_t numflag, dim_t n, INTL *ia, dim_t *ja, COEF *a, INTL **gia, dim_t **gja, COEF **ga)
{
  /****************************************************************************************/
  /* This function symmetrizes a triangular matrix                                        */
  /* On entry:                                                                            */
  /* flag_t job : =0 only the pattern is symmetrized (null entries are added)                */
  /*           =1 null entries are added in the matrix so that the pattern is symmetric   */
  /*           =2 the coefficient are also symmetrized  (B = A+At-Diag)                   */
  /* int numflag : =0 C numbering                                                         */
  /*               =1 Fortran numbering                                                   */                        
  /* n, ia, ja, a : the triangular matrix in CSR (or CSC) format                          */
  /*              : if a == NULL then only ia and ja are considered                       */
  /* On return:                                                                           */
  /* gia, gja, ga : the symmetrized matrix     TODO: nom                                  */
  /* NOTE : gia, gja, ga are allocated in this function                                   */
  /****************************************************************************************/
  
  INTL nnz;
#ifdef SPKIT_F 
  dim_t i;
#endif /* SPKIT_F */
  int j, ierr;
  INTL *ib;
  dim_t *jb, *iwk;
  flag_t job2;
  COEF *b = NULL;


 
#ifdef DEBUG_M
  assert(ia[0] == numflag);
#endif

#ifdef SPKIT_F 
  if(numflag == 0) {
    CSR_Cnum2Fnum(ja, ia, n);
  }
#else
  if(numflag == 1) {
    CSR_Fnum2Cnum(ja, ia, n);
  }
#endif

  if(job != 0)
    job2 = 1;
  else
    job2 = 0;

#ifdef SPKIT_F 
  nnz = ia[n]-1;
#else
  nnz = ia[n];
#endif


  if(nnz > 0)
    jb = (int *)malloc(nnz*sizeof(int));
  else
    jb = NULL;

  ib = (INTL *)malloc((n+1)*sizeof(INTL));
  if(job != 0)
    b = (COEF *)malloc(nnz*sizeof(COEF));
  
  iwk = (int *)malloc(n*sizeof(int));
  
  *gja = (dim_t *)malloc((2*nnz+1)*sizeof(dim_t));
  *gia = (INTL *)malloc((n+1)*sizeof(INTL));
  if(job != 0)
    *ga = (COEF *)malloc((2*nnz+1)*sizeof(COEF));


  /******** Symmetrize the pattern *****************/



#ifndef SPKIT_F
  csr2csc(n, job2, a, ja, ia, b, jb, ib);
#else
  i = 1;
  FFcsrcsc(&n, &job2, &i, a, ja, ia, b, jb, ib);

#ifdef DEBUG_M
  checkCSR(n, ib, jb, b, 1);
#endif

#endif

  if(job == 1 && nnz > 0)
    bzero(b, sizeof(COEF)*nnz);

  if(job == 2) /*** Annul the diagonal in At ****/
    {
      dim_t i;
      for(i=0;i<n;i++)
#ifdef SPKIT_F 
	for(j=ib[i]-1; j<ib[i+1]-1;j++)
	  if(jb[j] == i+1)
#else
	for(j=ib[i]; j<ib[i+1];j++)
	  if(jb[j] == i)
#endif
	    b[j] = 0.0;
    }

#ifdef SPKIT_F 
  i = 2*nnz-n;
#endif

#ifndef SPKIT_F
  if(job == 0) 
    ierr = aplb(n, n, job, NULL, ja, ia, NULL, jb, ib, NULL, *gja, *gia, 2*nnz+1/* -n? */, iwk);
  else {
    ierr = aplb(n, n, job, a, ja, ia, b, jb, ib, *ga, *gja, *gia, 2*nnz+1/* -n */, iwk);
  }
#else 
  if(job == 0)
    FFaplb(&n, &n, &job, NULL, ja, ia, NULL, jb, ib, NULL, *gja, *gia, &i, iwk, &ierr);
  else 
    FFaplb(&n, &n, &job, a, ja, ia, b, jb, ib, *ga, *gja, *gia, &i, iwk, &ierr);
#endif

  if(ierr != 0)
    {
      fprintf(stderr, "Error in aplb : ierr = %ld nnz = %ld \n", (long)ierr, (long)nnz);
      assert(ierr==0);
    }

#ifdef SPKIT_F
  nnz = (*gia)[n]-1;
#else
  nnz = (*gia)[n];
#endif

  (*gja) = (int *)realloc((*gja), nnz*sizeof(int));
  if(job != 0)
    (*ga) = (COEF *)realloc((*ga), nnz*sizeof(COEF));

  free(iwk);

  if(jb!=NULL)
    free(jb); 
  if(ib!=NULL)
    free(ib); 
  if(b!=NULL)
    free(b);
  
#ifdef SPKIT_F 
  if(numflag == 0)
    {
      CSR_Fnum2Cnum(ja, ia, n);
      CSR_Fnum2Cnum(*gja, *gia, n);
    }
#else
  if(numflag == 1)
    {
      CSR_Cnum2Fnum(ja, ia, n);
      CSR_Cnum2Fnum(*gja, *gia, n);
    }
#endif



}
