/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : find_MIS.c                                                 */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "phidal_ordering.h"
#include "queue.h"


void find_MIS(flag_t job, dim_t n, INTL *ia, dim_t *ja, INTL *sizetab, flag_t *mask, flag_t maskval, dim_t *MISsize, dim_t *MIS)
{
  /***************************************************************************************/
  /* This function compute an Maximum Independent Set                                    */
  /* On entry :                                                                          */
  /*      job       : if job = 0 try to maximize cardinal of the MIS                     */
  /*                  if job != 0 try to maximize weight of the MIS according to sizetab */
  /*      n, ia, ja : the graph of vertex in CSR format                                  */
  /*                  NB: NO EDGE from vertex i to vertex i are allowed                  */
  /*      sizetab   : sizetab[i] is the weight of vertex i                               */
  /*                  if sizetab == NULL all the vertice have the smae weight            */
  /*      mask      : if mask[i] != maskval then vertex i cannot be selected in the MIS  */
  /*                  if mask == NULL then all the vertice can be selected in the MIS    */
  /* On return :                                                                         */
  /*      MISsize   : the size of the MIS found                                          */
  /*      MIS       : the list of vertex i selected in the MIS                           */
  /*                  NB: MIS is allocated by user ; it must have a sufficient size      */
  /***************************************************************************************/
  
  int i,j, index;
  int v, neigh, nneigh;
  flag_t *lmask;
  INTL *lsizetab;
  int_t *degr;
  Queue heap;


  if(n==0)
    {
      (*MISsize) = 0;
      return;
    }

  lmask = (flag_t *)malloc(sizeof(flag_t)*n);
  if(mask == NULL)
    {
      maskval = 0;
      bzero(lmask, sizeof(flag_t)*n);
    }
  else
    memcpy(lmask, mask, sizeof(flag_t)*n);

  
  lsizetab = (INTL *)malloc(sizeof(INTL)*n);
  if(sizetab == NULL)
    bzero(lsizetab, sizeof(INTL)*n);
  else
    {
      /** We want to take sizetab into account with a precision of 10 percents compared to the average node size **/
      REAL average;
      average = 0;
      j = 0;
      for(i=0;i<n;i++)
	if(lmask[i] == maskval)
	  {
	    average += sizetab[i];
	    j++;
	  }

      average = fabs(average/j);

      memcpy(lsizetab, sizetab, sizeof(INTL)*n);

      for(i=0;i<n;i++)
	if(lmask[i] == maskval)
	  {
	    lsizetab[i] = (INTL)(lsizetab[i]*10.0/average);
	    /*fprintfv(5, stderr, "(i=%d sizetab = %d)  ", i, lsizetab[i]);*/
	  }
      /*fprintfv(5, stderr, "\n");*/
      
    }


  /***** Compute the degree (=number of unselected neighbor) for each vertex *****/
  degr = (int_t *)malloc(sizeof(int_t)*n);
  for(i=0;i<n;i++)
    {
      degr[i] = 0;
      for(j=ia[i];j<ia[i+1];j++)
	if(lmask[ja[j]] == maskval)
	  degr[i]++;
    }

  /**** Sort the vertice according to their degree and size ******/
  queueInit(&heap, 2*n);
  for(i=0;i<n;i++)
    if(lmask[i] == maskval)
      {
	if(job == 0)
	  queueAdd2(&heap, i, degr[i], -lsizetab[i]); 
	else
	  queueAdd2(&heap, i, -lsizetab[i], degr[i]); 
      }
  
  index = 0;
  while(queueSize(&heap) > 0)
    {
      v = queueGet(&heap);
      if(lmask[v] == maskval)
	{
	  lmask[v] = maskval+1;
	  MIS[index] = v;
	  index++;
	  /*** Mark all the neighbor of v and update degree ***/
	  for(i=ia[v];i<ia[v+1];i++)
	    {
	      neigh = ja[i];
	      if(lmask[neigh] == maskval)
		{
		  /*** Update the degree of the neighbor of neigh ***/
		  for(j=ia[neigh];j<ia[neigh+1];j++)
		    {
		      nneigh = ja[j];
		      if(lmask[nneigh] == maskval)
			{
			  degr[nneigh]--;
#ifdef DEBUG_M
			  assert(degr[nneigh] >= 0);
#endif
			}
		    }
		}
	    }

	  /**** Resort the vertice ****/
	   for(i=ia[v];i<ia[v+1];i++)
	    {
	      neigh = ja[i];
	      if(lmask[neigh] == maskval)
		{
		  lmask[neigh] = maskval+1; /** This vertex can not be selected in the MIS anymore ***/
		  /*** Update the degree of the neighbor of neigh ***/
		  for(j=ia[neigh];j<ia[neigh+1];j++)
		    {
		      nneigh = ja[j];
		      if(lmask[nneigh] == maskval)
			{
			  if(job == 0)
			    queueAdd2(&heap, nneigh, degr[nneigh], -lsizetab[nneigh]); 
			  else
			    queueAdd2(&heap, nneigh, -lsizetab[nneigh], degr[nneigh]); 
			}
		    }
		}
	    }

	}
      
    }
  (*MISsize) = index;
  
  /** Free memory ***/
  queueExit(&heap);
  free(degr);
  free(lmask);
  free(lsizetab);

}
