/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalDistrPrec_GEMM_ILUCT.c                               */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#define PARALLEL

#include "phidal_parallel.h"


void PhidalDistrPrec_GEMM_ILUCT(int_t levelnum, REAL *droptab, PhidalDistrMatrix *G, PhidalDistrMatrix *W, PhidalDistrMatrix *A, PhidalDistrPrec *P, PhidalDistrHID *BL, PhidalOptions *option)
{
  PhidalDistrMatrix *VL, *VU;
  PhidalMatrix *M;
  chrono_t t1, t2;
  dim_t i;


#ifdef DEBUG_M
  assert(option->schur_method != 1);
  assert(A->M.tli == A->M.tlj);
  assert(A->M.bri == A->M.brj);
  assert(G->M.tli == A->M.tli);
  assert(G->M.bri == A->M.bri);
  assert(G->M.brj == A->M.tlj-1);
  assert(W->M.tlj == A->M.tlj);
  assert(W->M.brj == A->M.brj);
  assert(W->M.bri == A->M.tli-1);
  assert(levelnum == option->forwardlev);
#endif

  t1 = dwalltime();
  
  P->dim = A->M.dim1;
  P->symmetric = option->symmetric;
  P->schur_method = 2;
  P->forwardlev = 0;
  P->levelnum =  levelnum;

 
  /** IMPORTANT: Cut U before L **/
  VU = (PhidalDistrMatrix *)malloc(sizeof(PhidalDistrMatrix));
  PhidalDistrMatrix_Init(VU);
  PhidalDistrMatrix_Setup(G->M.tlj, G->M.tlj, A->M.bri, A->M.bri, "U", "N", option->locally_nbr, VU, BL);
  PhidalDistrMatrix_Cut(A, VU, BL);


  VL = (PhidalDistrMatrix *)malloc(sizeof(PhidalDistrMatrix));
  PhidalDistrMatrix_Init(VL);
  PhidalDistrMatrix_Setup(G->M.tlj, G->M.tlj, A->M.bri, A->M.bri, "L", "N", option->locally_nbr, VL, BL);
  PhidalDistrMatrix_Cut(A, VL, BL);
  /*** Reset the block diagonal to null matrices **/
  M = &(VL->M);
  for(i=A->M.tli;i<=A->M.brj;i++)
    reinitCS(M->ca[ M->cia[i]]);

#ifdef DEBUG_M
  assert(G->M.dim2+A->M.dim2 == VL->M.dim2);
  assert(W->M.dim1+A->M.dim1 == VL->M.dim2);
#endif
  PhidalDistrMatrix_Cut(G, VL, BL);
  PhidalDistrMatrix_Cut(W, VU, BL);

  PhidalDistrMatrix_ILUCT_Restrict(A->M.tli, 0, VL, VU, option->droptol1, droptab, option->fillrat, BL, option, P->info);

  M_MALLOC(P->LU, D(SCAL));
  PREC_L_SCAL(P) = (PhidalDistrMatrix *)malloc(sizeof(PhidalDistrMatrix));
  PhidalDistrMatrix_Setup(A->M.tli, A->M.tli, A->M.bri, A->M.bri, "L", "N", option->locally_nbr, PREC_L_SCAL(P), BL);
  PhidalDistrMatrix_Cut(VL, PREC_L_SCAL(P), BL);
  PREC_U_SCAL(P) = (PhidalDistrMatrix *)malloc(sizeof(PhidalDistrMatrix));
  PhidalDistrMatrix_Setup(A->M.tli, A->M.tli, A->M.bri, A->M.bri, "U", "N", option->locally_nbr, PREC_U_SCAL(P), BL);
  PhidalDistrMatrix_Cut(VU, PREC_U_SCAL(P), BL);

  /*** OIMBE attention retirer VL de mem si corrige pas peak dans GEMM_I..T **/
  if(P->info != NULL)
    {
      /*fprintfd(stderr, "VLnnz = %ld \n", (long)PhidalDistrMatrix_NNZ(VL)); == 0*/
      PrecInfo_SubNNZ(P->info, PhidalDistrMatrix_NNZ(VL));
      /*fprintfd(stderr, "VUnnz = %ld \n", (long)PhidalDistrMatrix_NNZ(VU)); == 0*/
      PrecInfo_SubNNZ(P->info, PhidalDistrMatrix_NNZ(VU));
    }
  PhidalDistrMatrix_Clean(VL);
  PhidalDistrMatrix_Clean(VU);
  free(VL);
  free(VU);

  t2 = dwalltime();
  fprintfv(5, stderr, "Proc %d: Time in GEMM_ILUCT = %g \n", BL->proc_id, t2-t1);
}
