/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalMatrix_solve.c                                       */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_sequential.h"

void PhidalMatrix_Lsolve(flag_t unitdiag, PhidalMatrix *L, COEF *x, COEF *b, PhidalHID *BL)
{
  /*****************************************************************************************/
  /* This function compute x = L^-1.b                                                      */
  /* If unitdiag == 1 then the diagonal is considered to be unitary (stored or not stored) */
  /*   THIS CONCERNS THE DIAGONAL NOT THE BLOCK DIAGONAL                                   */
  /* Work also in place (i.e. x == b)                                                      */
  /*****************************************************************************************/
  dim_t i, j, k;
  int *bind;
  COEF *xptr;
  int offseti;
#ifdef DEBUG_M
  assert(L->dim1 == L->dim2);
  assert(L->tli == L->tlj);
  assert(L->bri == L->brj);
#endif

  bind = BL->block_index;
  offseti = bind[L->tli];

  if(x != b)
    memcpy(x, b, sizeof(COEF)*L->dim1);

  if(L->csc == 0)
    for(i=L->tli;i<=L->bri;i++)
      {
	xptr = x+(bind[i]-offseti);
	for(k=L->ria[i];k<L->ria[i+1]-1;k++)
	  {
	    j = L->rja[k];
	    matvecz(L->ra[k], x+(bind[j]-offseti), xptr, xptr); 
	  }
#ifdef DEBUG_M
	assert(L->rja[L->ria[i+1]-1] == i);
#endif
	CSR_Lsol(unitdiag, L->ra[L->ria[i+1]-1], xptr, xptr);
      }
  else
    for(i=L->tli;i<=L->bri;i++)
      {
	xptr = x+(bind[i]-offseti);
	for(k=L->ria[i];k<L->ria[i+1]-1;k++)
	  {
	    j = L->rja[k];
	    CSC_matvecz(L->ra[k], x+(bind[j]-offseti), xptr, xptr); 
	  }
#ifdef DEBUG_M
	assert(L->rja[L->ria[i+1]-1] == i);
#endif
	CSC_Lsol(unitdiag, L->ra[L->ria[i+1]-1], xptr, xptr);
      }
}


void PhidalMatrix_Usolve(flag_t unitdiag, PhidalMatrix *U, COEF *x, COEF *b, PhidalHID *BL)
{
  /*********************************************************/
  /* This function compute x = U^-1.b                      */
  /* Work also in place (i.e. x == b)                      */
  /* If unitdiag == 1 then the diagonal is considered to 
     be unitary (stored or not stored) */
  /*   THIS CONCERNS THE DIAGONAL NOT THE BLOCK DIAGONAL   */
  /*********************************************************/
  dim_t i, j, k;
  int *bind;
  COEF *xptr;
  int offseti;
#ifdef DEBUG_M
  assert(U->dim1 == U->dim2);
  assert(U->tli == U->tlj);
  assert(U->bri == U->brj);
#endif

  bind = BL->block_index;
  offseti = bind[U->tli];

  if(x != b)
    memcpy(x,b, sizeof(COEF)*U->dim1);

  if(U->csc == 0)
    for(j=U->brj;j>=U->tlj;j--)
      {
	xptr = x+(bind[j]-offseti);
#ifdef DEBUG_M
	assert(U->cja[U->cia[j+1]-1] == j);
#endif
	
	CSR_Usol(unitdiag, U->ca[U->cia[j+1]-1], xptr, xptr);
	
	for(k=U->cia[j+1]-2;k>=U->cia[j];k--)
	  {
	    i = U->cja[k];
	    matvecz(U->ca[k], xptr, x+(bind[i]-offseti), x+(bind[i]-offseti)); 
	  }

      }
  else
    for(j=U->brj;j>=U->tlj;j--)
      {
	xptr = x+(bind[j]-offseti);
#ifdef DEBUG_M
	assert(U->cja[U->cia[j+1]-1] == j);
#endif
	
	CSC_Usol(unitdiag, U->ca[U->cia[j+1]-1], xptr, xptr);
	
	for(k=U->cia[j+1]-2;k>=U->cia[j];k--)
	  {
	    i = U->cja[k];
	    CSC_matvecz(U->ca[k], xptr, x+(bind[i]-offseti), x+(bind[i]-offseti)); 
	  }
	
      }
}


